
#include <stdio.h>
#include "windef.h"
#include "globdef.h"
#include "uidef.h"
#include "fft1def.h"
#include "fft2def.h"
#include "fft3def.h"
#include "screendef.h"
#include "sigdef.h"
#include "seldef.h"
#include "blnkdef.h"
#include "caldef.h"
#include "sdrdef.h"
#include "txdef.h"
#include "vernr.h"
#include "keyboard_def.h"
#include "thrdef.h"
#include "hwaredef.h"
#include "ftdi.h"

sdrWrite P_Write;
sdrRead P_Read;
sdrClose P_Close;
sdrSetTimeouts P_SetTimeouts;
FT_HANDLE sdr14_handle;

void lir_sdr14_write(void *s, int bytes)
{
DWORD written;
P_Write(sdr14_handle,s,bytes, &written);
}

int lir_sdr14_read(void *s,int bytes)
{
DWORD i;
P_Read(sdr14_handle, s, bytes, &i);
return i;
}

void lir_close_sdr14(void)
{
sdr14_stop();
P_Close(sdr14_handle);
sdr=-1;
}


void open_sdr14(void)
{
sdrOpenEx P_OpenEx;
HINSTANCE h_sdr14;
h_sdr14 = LoadLibrary("ftd2xx.dll");
if (h_sdr14 == NULL) goto open_sdr14_fail;
P_OpenEx=(sdrOpenEx)GetProcAddress(h_sdr14, "FT_OpenEx");
if(P_OpenEx == NULL)goto open_sdr14_fail;
if(P_OpenEx(SDR14_NAME_STRING, FT_OPEN_BY_DESCRIPTION, 
                              &sdr14_handle) != FT_OK)goto open_sdr14_fail;
P_Read=(sdrRead)GetProcAddress(h_sdr14, "FT_Read");
if(P_Read == NULL)goto open_sdr14_fail;
P_Write = (sdrWrite)GetProcAddress(h_sdr14, "FT_Write");
if(P_Write == NULL)goto open_sdr14_fail;
P_Close = (sdrClose)GetProcAddress(h_sdr14, "FT_Close");
if(P_Close == NULL)goto open_sdr14_fail;
P_SetTimeouts = (sdrSetTimeouts)GetProcAddress(h_sdr14, "FT_SetTimeouts");
if(P_SetTimeouts == NULL)goto open_sdr14_fail;
if(P_SetTimeouts(sdr14_handle,500,500) != FT_OK)goto open_sdr14_fail;
sdr=1;
return;  
open_sdr14_fail:;
FreeLibrary(h_sdr14);
sdr=-1;
}

DWORD WINAPI winthread_rx_adinput(PVOID arg)
{
int timing_loop_counter,timing_loop_counter_max,initial_skip_flag;
int local_rxadin1_newbuf_ptr;
int local_rxadin2_newbuf_ptr;
double dt1, read_start_time,total_reads;
double total_time2;
MMRESULT mmrs;
WAVEHDR *whdr1, *whdr2;
char *c1, *c2;
short int *c1_isho, *c2_isho;
int *c1_int, *c2_int, *rxin_int; 
int i, k, screen_loop_counter, screen_loop_counter_max;
c1=arg;
open_rx_sndin();
if(kill_all_flag) goto rxadin_init_error;
SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_HIGHEST);
thread_status_flag[THREAD_RX_ADINPUT]=THRFLAG_ACTIVE;
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
local_rxadin1_newbuf_ptr=rxadin1_newbuf_ptr;
local_rxadin2_newbuf_ptr=rxadin2_newbuf_ptr;
screen_loop_counter_max=0.1*interrupt_rate;
if(screen_loop_counter_max==0)screen_loop_counter_max=1;
screen_loop_counter=interrupt_rate;
timing_loop_counter_max=interrupt_rate;
timing_loop_counter=2.5*timing_loop_counter_max;
read_start_time=current_time();
total_reads=0;
initial_skip_flag=1;
while(thread_command_flag[THREAD_RX_ADINPUT] == THRFLAG_ACTIVE)
  {
  timing_loop_counter--;
  if(timing_loop_counter == 0)
    {
    timing_loop_counter=timing_loop_counter_max;
    total_time2=current_time();
    if(initial_skip_flag != 0)
      {
      read_start_time=total_time2;
      total_reads=0;
      initial_skip_flag=0;
      }
    else
      {
      total_reads+=timing_loop_counter_max;
      dt1=total_time2-read_start_time;
      measured_ad_speed=total_reads*ad_read_fragments/dt1;
      }
    }
  if(hware_flag != 0)
    {    
    dt1=current_time();
    if(dt1 - hware_time > 0.005)
      {
      control_hware();
      hware_time=dt1;
      }
    }
// Here we post to the screen routine every 0.1 second.      
  screen_loop_counter--;
  if(screen_loop_counter == 0)
    {
    screen_loop_counter=screen_loop_counter_max;
    lir_sem_post(SEM_SCREEN);
    }
  WaitForSingleObject(rxin1_bufready, INFINITE);
  rxin_isho=(void*)(&timf1_char[timf1p_pa]);
  timf1p_pa=(timf1p_pa+ad_read_bytes)&timf1_bytemask;
  whdr1=(WAVEHDR*)rxadin1_newbuf[local_rxadin1_newbuf_ptr];
  if(local_rxadin1_newbuf_ptr==rxadin1_newbuf_ptr)
    {
    lirerr(1224);
    goto rxadin_error;
    }
  if( (whdr1[0].dwFlags & WHDR_DONE) == 0)  
    {
    lirerr(1225);
    goto rxadin_error;
    }
  c1=whdr1[0].lpData;  
  if(ui.ad_device_no < 256)
    {
    memcpy(rxin_isho,c1,ad_read_bytes);
    local_rxadin1_newbuf_ptr++;
    if(local_rxadin1_newbuf_ptr>=no_of_rx_wavein)local_rxadin1_newbuf_ptr=0;
    mmrs=waveInAddBuffer(hwav_rxadin1,whdr1, sizeof(WAVEHDR));
    if(mmrs != MMSYSERR_NOERROR)
      {
      lirerr(1226);
      goto rxadin_error;
      }
    finish_rx_read(rxin_isho);
    if(kill_all_flag) goto rxadin_error;
    }
  else
    {
    WaitForSingleObject(rxin2_bufready, INFINITE);
    whdr2=(WAVEHDR*)rxadin2_newbuf[local_rxadin2_newbuf_ptr];
    if(local_rxadin2_newbuf_ptr==rxadin2_newbuf_ptr)
      {
      lirerr(1230);
      goto rxadin_error;
      }
    if( (whdr2[0].dwFlags & WHDR_DONE) == 0)  
      {
      lirerr(1231);
      goto rxadin_error;
      }
    c2=whdr2[0].lpData;  
    if( (ui.input_mode&DWORD_INPUT) == 0)
      {
      k=ad_read_bytes>>3;
      c1_isho=(void*)whdr1[0].lpData;  
      c2_isho=(void*)whdr2[0].lpData;  
      for(i=0; i<k; i++)
        {
        rxin_isho[4*i  ]=c1_isho[2*i  ];          
        rxin_isho[4*i+1]=c1_isho[2*i+1];          
        rxin_isho[4*i+2]=c2_isho[2*i  ];          
        rxin_isho[4*i+3]=c2_isho[2*i+1];          
        }
      }
    else
      {
      k=ad_read_bytes>>4;
      c1_int=(void*)whdr1[0].lpData;  
      c2_int=(void*)whdr2[0].lpData;  
      rxin_int=(void*)rxin_isho;
      for(i=0; i<k; i++)
        {
        rxin_int[4*i  ]=c1_int[2*i  ];          
        rxin_int[4*i+1]=c1_int[2*i+1];          
        rxin_int[4*i+2]=c2_int[2*i  ];          
        rxin_int[4*i+3]=c2_int[2*i+1];          
        }
      }
    local_rxadin1_newbuf_ptr++;
    local_rxadin2_newbuf_ptr++;
    if(local_rxadin1_newbuf_ptr>=no_of_rx_wavein)local_rxadin1_newbuf_ptr=0;
    if(local_rxadin2_newbuf_ptr>=no_of_rx_wavein)local_rxadin2_newbuf_ptr=0;
    mmrs=waveInAddBuffer(hwav_rxadin1,whdr1, sizeof(WAVEHDR));
    if(mmrs != MMSYSERR_NOERROR)
      {
      lirerr(1232);
      goto rxadin_error;
      }
    mmrs=waveInAddBuffer(hwav_rxadin2,whdr2, sizeof(WAVEHDR));
    if(mmrs != MMSYSERR_NOERROR)
      {
      lirerr(1233);
      goto rxadin_error;
      }
    finish_rx_read(rxin_isho);
    if(kill_all_flag) goto rxadin_error;
    }
  }
rxadin_error:;
close_rx_sndin();
rxadin_init_error:;
thread_status_flag[THREAD_RX_ADINPUT]=THRFLAG_RETURNED;
if(!kill_all_flag)
  {
  while(thread_command_flag[THREAD_RX_ADINPUT] != THRFLAG_NOT_ACTIVE)
    {
    lir_sleep(1000);
    }
  }  
return 100;
}



int lir_output_bytes(void)
{
// Find out how much empty space we have in the output buffers.
WAVEHDR *whdr;
int i,k;
if(wave_out_open_flag)
  {
  k=0;
  for(i=0; i<NO_OF_RX_WAVEOUT; i++)
    {
    whdr=rxdaout_newbuf[i];
    if( (whdr[0].dwFlags&WHDR_INQUEUE) == 0)k++; 
    }
  }
else
  {
  k=NO_OF_RX_WAVEOUT;
  }
return k*daout_block;
}


int make_da_wts(void)
{
int i, k;
da_totbytes=NO_OF_RX_WAVEOUT*daout_block;
k=lir_output_bytes();
if(k > da_maxbytes)da_maxbytes=k;
i=da_totbytes-k;
i/=(da_output_channels*da_output_bytes);
return i;
}

void lir_empty_da_device_buffer(void)
{
int i;
if(wave_out_open_flag)
  {
  if(ui.da_stopstart == TRUE)
    {
    close_rx_sndout();
    open_rx_sndout();
    }
  else
    {  
    waveOutReset(hwav_rxdaout);
    }
  }
else   
  {
  open_rx_sndout();
  if(kill_all_flag) return;
  for(i=0; i<daout_block; i++)da_wrbuf[i]=0;
  lir_dawrite();
  if(kill_all_flag) return;
  waveOutRestart(hwav_rxdaout);
  }
}

void close_rx_sndout(void)
{
MMRESULT mmrs;
WAVEHDR *whdr;
int i;
if(!wave_out_open_flag)return;
mmrs=waveOutReset(hwav_rxdaout);
if(mmrs != MMSYSERR_NOERROR)lirerr(25344);
for(i=0; i<NO_OF_RX_WAVEOUT; i++)
  {
  whdr=rxdaout_newbuf[i];
  while( (whdr[0].dwFlags&WHDR_INQUEUE) != 0)lir_sleep(3000); 
  mmrs=waveOutUnprepareHeader(hwav_rxdaout, whdr, sizeof(WAVEHDR));
  if(mmrs != MMSYSERR_NOERROR)lirerr(25144);
  }
mmrs=waveOutClose(hwav_rxdaout);
if(mmrs != MMSYSERR_NOERROR)lirerr(25344);
free(da_wrbuf);
free(wave_outhdr);
free(waveout_buf);
wave_out_open_flag=FALSE;
audio_out=-1;
}

void lir_dawrite(void)
{
char *c;
WAVEHDR *whdr;
// We arrive here because one more block of data is available
// in da_wrbuf.
// Wait until there is a buffer ready for us.
whdr=rxdaout_newbuf[rxdaout_newbuf_ptr];
while( (whdr[0].dwFlags&WHDR_INQUEUE) != 0) 
  {
  lir_sleep(10000);
  }
c=whdr[0].lpData;  
memcpy(c,da_wrbuf,daout_block);
waveOutWrite(hwav_rxdaout,whdr,sizeof(WAVEHDR));
rxdaout_newbuf_ptr++;
if(rxdaout_newbuf_ptr >= NO_OF_RX_WAVEOUT)rxdaout_newbuf_ptr=0;
}

DWORD WINAPI winthread_rx_output(PVOID arg)
{
char *c;
c=arg;
// The rx_output routine will call lir_dawrite() every time there is
// enough data available to fill one buffer.
SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_HIGHEST);
rx_output();
return 100;
}


void close_rx_sndin(void)
{
if(wave_in_open_flag1 == 0)return;
if(ui.ad_device_no > 255)
  {
  waveInReset(hwav_rxadin2);
  waveInClose(hwav_rxadin2);
  }
waveInReset(hwav_rxadin1);
waveInClose(hwav_rxadin1);
while(wave_in_open_flag1!=0||wave_in_open_flag2!=0)
  {
  lir_sleep(1000);
  }
free(wave_inhdr);
free(wavein_buf);
}

VOID CALLBACK rx_daout(HWAVEOUT hwi, UINT iMsg, DWORD dwInst, 
                         DWORD dwParam1, DWORD dwParam2)
{
DWORD i;
i=(DWORD)hwi;
i=dwInst;
i=dwParam2;
i=iMsg;
i=dwParam1;
if(iMsg == WIM_DATA)
  {
  }
else
  {
  if(iMsg == WIM_CLOSE)
    {
    i=(DWORD)hwi;
    i=dwInst;
    i=dwParam2;
    audio_in=-1;
    }
  else
    {
    if(iMsg == WIM_OPEN)
      {
      audio_in=0;
      }
    }  
  }
}                         


VOID CALLBACK rx_adin1(HWAVEIN hwi, UINT iMsg, DWORD dwInst, 
                         DWORD dwParam1, DWORD dwParam2)
{
long old;
DWORD i;
if(iMsg == WIM_DATA)
  {
  rxadin1_newbuf[rxadin1_newbuf_ptr]=dwParam1;
  rxadin1_newbuf_ptr++;
  if(rxadin1_newbuf_ptr >= no_of_rx_wavein)rxadin1_newbuf_ptr=0;
  ReleaseSemaphore(rxin1_bufready,1,&old);
  }
else
  {
  if(iMsg == WIM_CLOSE)
    {
    i=(DWORD)hwi;
    i=dwInst;
    i=dwParam2;
    wave_in_open_flag1=0;
    }
  }
}                         


VOID CALLBACK rx_adin2(HWAVEIN hwi, UINT iMsg, DWORD dwInst, 
                         DWORD dwParam1, DWORD dwParam2)
{
long old;
DWORD i;
if(iMsg == WIM_DATA)
  {
  rxadin2_newbuf[rxadin2_newbuf_ptr]=dwParam1;
  rxadin2_newbuf_ptr++;
  if(rxadin2_newbuf_ptr >= no_of_rx_wavein)rxadin2_newbuf_ptr=0;
  ReleaseSemaphore(rxin2_bufready,1,&old);
  }
else
  {
  if(iMsg == WIM_CLOSE)
    {
    i=(DWORD)hwi;
    i=dwInst;
    i=dwParam2;
    wave_in_open_flag2=0;
    }
  }
}                         



void open_rx_sndin(void)
{
int errcod;
WAVEFORMATEX fmt;
MMRESULT mmrs;
int i, k, dev, devno;
if(wave_in_open_flag1 == 1)
  {
  lirerr(869233);
  return;
  }
if( disksave_flag == 2 || 
    network_flag == 2 ||
    ui.ad_device_no == SDR14_DEVICE_CODE )
  {
  return;
  }
dev=ui.ad_device_no&255;
fmt.nChannels=ui.ad_channels;
if(ui.ad_device_no > 255)
  {
  devno=2;
  fmt.nChannels/=2;
  }
else
  {
  devno=1;
  }
fmt.wFormatTag=WAVE_FORMAT_PCM;
fmt.nSamplesPerSec=ui.input_speed;
if( (ui.input_mode&DWORD_INPUT)==0)
  {
  fmt.wBitsPerSample=16;
  }
else
  {
  fmt.wBitsPerSample=32;
  }  
fmt.cbSize=0;
fmt.nBlockAlign=fmt.nChannels*fmt.wBitsPerSample/8;
fmt.nAvgBytesPerSec=fmt.nBlockAlign*fmt.nSamplesPerSec;
mmrs=waveInOpen(&hwav_rxadin1, dev, &fmt, (DWORD)rx_adin1, 0, CALLBACK_FUNCTION);
if(mmrs != MMSYSERR_NOERROR)
  {
  if(mmrs == MMSYSERR_ALLOCATED)
    {
    lirerr(1192);
    }
  else
    {
    lirerr(1227);
    }
  return;
  }
wave_in_open_flag1=1;
rxin1_bufready=CreateSemaphore( NULL, 0, 1024, NULL);
if(rxin1_bufready== NULL)
  {
  errcod=1223;
  goto errclose;
  }    
if(devno==2)
  {
  dev=(ui.ad_device_no>>8)-1;
  mmrs=waveInOpen(&hwav_rxadin2, dev, &fmt, (DWORD)rx_adin2, 0, CALLBACK_FUNCTION);
  if(mmrs != MMSYSERR_NOERROR)
    {
    errcod=1228;
    goto errclose;
    }
  wave_in_open_flag2=1;
  rxin2_bufready=CreateSemaphore( NULL, 0, 1024, NULL);
  if(rxin2_bufready== NULL)
    {
    errcod=1229;
    goto errclose;
    }    
  }
// We read ui.input_speed*bytes_per_fragment bytes per second from hardware.
// The callback routine will give us the data in a block size of 
// ad_read_bytes/devno.
// Make the number of buffers large enough to hold 0.05 seconds of data.
no_of_rx_wavein=0.05*(float)(ui.input_speed*bytes_per_fragment)/ad_read_bytes;
if(no_of_rx_wavein<4)no_of_rx_wavein=4;
wavein_buf=malloc(no_of_rx_wavein*ad_read_bytes+fmt.nBlockAlign);
if(wavein_buf == NULL)
  {
  errcod=1236;
  goto errclose;
  }
wave_inhdr=malloc(devno*no_of_rx_wavein*(sizeof(WAVEHDR)+sizeof(DWORD)));
if(wave_inhdr == NULL)
  {
  errcod=1236;
  goto errfree_inbuf;
  }
rxadin1_newbuf=(void*)&wave_inhdr[devno*no_of_rx_wavein];
rxadin2_newbuf=&rxadin1_newbuf[no_of_rx_wavein];
rxadin1_newbuf_ptr=0;
rxadin2_newbuf_ptr=0;
for(i=0; i<no_of_rx_wavein; i++)
  {
  if(devno==1)
    {
    k=(int)(wavein_buf+i*ad_read_bytes+fmt.nBlockAlign);
    k&=(-1-fmt.nBlockAlign);
    wave_inhdr[i].lpData=(char*)(k);
    wave_inhdr[i].dwBufferLength=ad_read_bytes;
    wave_inhdr[i].dwBytesRecorded=0;
    wave_inhdr[i].dwUser=0;
    wave_inhdr[i].dwFlags=0;
    wave_inhdr[i].dwLoops=0;
    wave_inhdr[i].lpNext=NULL;
    wave_inhdr[i].reserved=0;
    mmrs=waveInPrepareHeader(hwav_rxadin1, &wave_inhdr[i], sizeof(WAVEHDR));
    if(mmrs != MMSYSERR_NOERROR)
      {
      errcod=1245;
      goto errfree;
      }
    mmrs=waveInAddBuffer(hwav_rxadin1,&wave_inhdr[i], sizeof(WAVEHDR));
    if(mmrs != MMSYSERR_NOERROR)
      {
      errcod=1247;
      goto errfree;
      }
    }
  else
    {
    k=(int)(wavein_buf+(2*i  )*(ad_read_bytes>>1)+fmt.nBlockAlign);
    k&=(-1-fmt.nBlockAlign);
    wave_inhdr[2*i  ].lpData=(char*)(k);
    wave_inhdr[2*i  ].dwBufferLength=ad_read_bytes>>1;
    wave_inhdr[2*i  ].dwBytesRecorded=0;
    wave_inhdr[2*i  ].dwUser=0;
    wave_inhdr[2*i  ].dwFlags=0;
    wave_inhdr[2*i  ].dwLoops=0;
    wave_inhdr[2*i  ].lpNext=NULL;
    wave_inhdr[2*i  ].reserved=0;
    mmrs=waveInPrepareHeader(hwav_rxadin1, &wave_inhdr[2*i  ], sizeof(WAVEHDR));
    if(mmrs != MMSYSERR_NOERROR)
      {
      errcod=1245;
      goto errfree;
      }
    mmrs=waveInAddBuffer(hwav_rxadin1,&wave_inhdr[2*i  ], sizeof(WAVEHDR));
    if(mmrs != MMSYSERR_NOERROR)
      {
      errcod=1247;
      goto errfree;
      }
    k=(int)(wavein_buf+(2*i+1)*(ad_read_bytes>>1)+fmt.nBlockAlign);
    k&=(-1-fmt.nBlockAlign);
    wave_inhdr[2*i+1].lpData=(char*)(k);
    wave_inhdr[2*i+1].dwBufferLength=ad_read_bytes>>1;
    wave_inhdr[2*i+1].dwBytesRecorded=0;
    wave_inhdr[2*i+1].dwUser=0;
    wave_inhdr[2*i+1].dwFlags=0;
    wave_inhdr[2*i+1].dwLoops=0;
    wave_inhdr[2*i+1].lpNext=NULL;
    wave_inhdr[2*i+1].reserved=0;
    mmrs=waveInPrepareHeader(hwav_rxadin2, &wave_inhdr[2*i+1], sizeof(WAVEHDR));
    if(mmrs != MMSYSERR_NOERROR)
      {
      errcod=1245;
      goto errfree;
      }
    mmrs=waveInAddBuffer(hwav_rxadin2,&wave_inhdr[2*i+1], sizeof(WAVEHDR));
    if(mmrs != MMSYSERR_NOERROR)
      {
      errcod=1247;
      goto errfree;
      }
    }        
  }  
waveInStart(hwav_rxadin1);
if(devno==2)waveInStart(hwav_rxadin2);
return;
errfree:;
free(wave_inhdr);
errfree_inbuf:;
free(wavein_buf);
errclose:;
if(devno==2)
  {
  waveInClose(hwav_rxadin2);
  }
waveInClose(hwav_rxadin1);
lirerr(errcod);
}





void set_analog_io(void)
{
FILE *sdr14_file;
char dev_channels[MAX_WAV_DEVICES];
char s[80], ss[80];
int *sdr_pi;
int i, j, k, dev, line;
float t1;
unsigned int device_id, no_of_addev, no_of_dadev;
MMRESULT mmrs;
WAVEINCAPS pwic;
WAVEOUTCAPS pwoc;
WAVEFORMATEX fmt;
ad_read_bytes=4096;
j=0;
line=1;
ui.ad_device_no=-1;
ui.da_device_no=-1;
clear_screen();
// Set device_no to appropriate values if some dedicated hardware
// is selected by the user.
//  ------------------------------------------------------------
// See if there is an SDR-14 on the system.
settextcolor(12);
lir_text(10,10,"Looking for an SDR-14.");
lir_text(10,11,"Reset hardware (CPU/USB and SDR-14) if system hangs here.");
open_sdr14();
settextcolor(7);
clear_lines(10,11);
if(sdr != -1)
  {
  lir_text(5,5,"An SDR-14 is detected.");
  lir_text(5,6,"Do you want to use it for RX input (Y/N)?");
qsdr:;
  await_processed_keyboard();
  if(kill_all_flag) goto sdr14_errexit;
  if(lir_inkey == 'Y')
    {
    ui.ad_device_no=SDR14_DEVICE_CODE;
    ui.input_mode=IQ_DATA;
    ui.rx_channels=1;
    ui.ad_channels=2;
    ui.ad_device_mode=0;
    sdr_target_name(ss);
    sprintf(s,"Target name: %s",ss);
    lir_text(5,10,s);
    if(strcmp(sdr14_name_string,ss) != 0)
      {
      lir_text(5,12,"Unknown hardware, can not use.");
      lir_text(5,13,press_any_key);
      await_keyboard();
      if(kill_all_flag) goto sdr14_errexit;
      clear_screen();  
      ui.ad_device_no=-1;
      }
    else
      {
      sdr_target_serial_number(ss);
      sprintf(s,"Serial no: %s",ss);
      lir_text(5,11,s);
      sdr_target_interface_version(ss);
      i=(short int)ss[0];
      sprintf(s,"Interface version: %d.%02d",i/100,i%100);
      lir_text(5,12,s);
      sdr_target_boot_version(ss);
      i=(short int)ss[0];
      sprintf(s,"PIC boot version: %d.%02d",i/100,i%100);
      lir_text(5,13,s);
      sdr_target_firmware_version(ss);
      i=(short int)ss[0];
      sprintf(s,"PIC firmware version: %d.%02d",i/100,i%100);
      lir_text(5,14,s);
      lir_text(10,16,"PRESS ANY KEY");
      await_processed_keyboard();
      if(kill_all_flag) goto sdr14_errexit;
      clear_screen();
      lir_text(10,line,"SDR-14 selected for input");
      line++;
      lir_text(5,line,"Set CIC2 decimation (2 - 16)");
      sdr14.m_cic2=lir_get_integer(35, line, 2, 2,16);
      if(kill_all_flag)goto sdr14_errexit;
      t1=SDR14_SAMPLING_CLOCK/sdr14.m_cic2;
      sprintf(s,"clk=%.2f MHz",t1);
      lir_text(39,line,s);
      line++;
      lir_text(5,line,"Set CIC5 decimation (2 - 32)");
      sdr14.m_cic5=lir_get_integer(35, line, 2, 2,32);
      if(kill_all_flag)goto sdr14_errexit;
      t1/=sdr14.m_cic5;
      sprintf(s,"clk=%.4f MHz",t1);
      lir_text(39,line,s);
      line++;
      lir_text(5,line,"Set RCF decimation (2 - 32)");
      sdr14.m_rcf=lir_get_integer(35, line, 2, 2,32);
      if(kill_all_flag)goto sdr14_errexit;
      t1/=sdr14.m_rcf;
      ui.input_speed=t1*1000000;
      sprintf(s,"clk=%.2f kHz",t1*1000);
      lir_text(39,line,s);
      line++;
      lir_text(5,line,"Set RCF output shift (0 - 7)");
      sdr14.ol_rcf=lir_get_integer(35, line, 2, 2,32);
      if(kill_all_flag)goto sdr14_errexit;
      line++;
      lir_text(5,line,"Select direct input (0 - 1)");
      sdr14.input=lir_get_integer(35, line, 2, 0, 1);
      if(kill_all_flag)goto sdr14_errexit;
      sdr14_file=fopen("par_sdr14","w");
      if(sdr14_file == NULL)
        {
        lirerr(381264);
sdr14_errexit:;
        lir_close_sdr14();
        return;
        }
      sdr14.check=SDR14PAR_VERNR;
      sdr_pi=(void*)(&sdr14);
      for(i=0; i<MAX_SDR14_PARM; i++)
        {
        fprintf(sdr14_file,"%s [%d]\n",sdr14_parm_text[i],sdr_pi[i]);
        }
      parfile_end(sdr14_file);
      }
    }
  else  
    {
    if(lir_inkey != 'N')goto qsdr;
    clear_screen();
    }
  lir_close_sdr14();
  }

//  ------------------------------------------------------------
if(ui.ad_device_no == -1)
  {
// look for other hardwares
  }  
if(ui.ad_device_no != -1)
  {
  if(ui.da_device_no == -1)
    {
    goto set_sndout;
    }
  else
    {
    return;
    }
  }    
// If none of the supported SDR hardwares is found, use the 
// sound cards for input and output.
// ----------------------------------------------------------
// Find out how many audio input devices there are on this computer.
// Allow the user to select one or two for the rx input.
settextcolor(14);
lir_text(20,0,"Available soundcards.");
settextcolor(7);
no_of_addev=waveInGetNumDevs();
if(no_of_addev == 0)
  {
  lir_text(5,2,"No soundcard detected.");
  lir_text(5,3,"PRESS ANY KEY");
  await_keyboard();
  return;
  }
line=2;
if(no_of_addev > MAX_WAV_DEVICES)no_of_addev=MAX_WAV_DEVICES;
for(device_id=0; device_id<no_of_addev; device_id++)
  {
  mmrs=waveInGetDevCaps(device_id, &pwic, sizeof(WAVEINCAPS));   
  if(mmrs==MMSYSERR_NOERROR)
    {
// The pwic.wChannels is filled correctly under Win98 only.
// Under Windows 2000 and Windows XP this parameter is always 65535 (-1)
// Check for availability of any reasonable format.
// WAVE_FORMAT_1M16   11.025 kHz, mono, 16-bit
// WAVE_FORMAT_2M16   22.05  kHz, mono, 16-bit
// WAVE_FORMAT_4M16   44.1   kHz, mono, 16-bit
// WAVE_FORMAT_1S16   11.025 kHz, stereo, 16-bit
// WAVE_FORMAT_2S16   22.05  kHz, stereo, 16-bit
// WAVE_FORMAT_4S16   44.1   kHz, stereo, 16-bit
    i=0;
    if( (pwic.dwFormats & 
           (WAVE_FORMAT_1M16|WAVE_FORMAT_2M16|WAVE_FORMAT_4M16)) != 0)i=1;
    if( (pwic.dwFormats & 
           (WAVE_FORMAT_1S16|WAVE_FORMAT_2S16|WAVE_FORMAT_4S16)) != 0)i|=2;
    dev_channels[device_id]=i;
    sprintf(s,"%d  Fmts=%4lx   %s", device_id, pwic.dwFormats, pwic.szPname);
    lir_text(10,line,s);
    line++;
    }
  else
    {
    dev_channels[device_id]=-1;
    }      
  }
line+=2;  
if(no_of_addev == 1)
  {
  ui.ad_device_no=0;
  j=dev_channels[0];
  if(j>2)j=2;
  }
else  
  {
  lir_text(0,line,"Select (first) device for Rx input by line number:");
get_dev1:;
  ui.ad_device_no=lir_get_integer(51,line, 2, 0, no_of_addev-1);
  if(kill_all_flag) return;
  if(dev_channels[ui.ad_device_no]<0)goto get_dev1;
  j=dev_channels[ui.ad_device_no];
  if(j>2)j=2;
  line++;
  lir_text(0,line,"Do you need more channels from the same soundcard ? (Y/N)");
get_dev2:;
  toupper_await_keyboard();
  if(kill_all_flag) return;
  if(lir_inkey == 'Y')
    {
    line++;
    lir_text(0,line,"Select a second device for Rx input:");
dev2_n:;
    i=lir_get_integer(37,line, 2, 0, no_of_addev-1);
    if(kill_all_flag) return;
    if(i==ui.ad_device_no || dev_channels[i]<0)goto dev2_n;
    ui.ad_device_no+=256*(i+1);
    j*=2;
    }
  else
    {
    if(lir_inkey != 'N')goto get_dev2;
    }
  }
// Windows does not allow us to query the device for its capabilities.
// Ask the user to specify what he wants.
line+=2;
lir_text(0,line,"Can not query hardware under Windows. (do not know how)");
line++;
lir_text(0,line,"Make sure you enter data compatible with your hardware");
line++;
lir_text(0,line,"native capabilities.");
line+=2;
get_rxinrate:;
lir_text(10,line,"Sampling speed:");
ui.input_speed=lir_get_integer(26,line, 6, 5000, 999999);
if(kill_all_flag) return;
get_rxbits:;
lir_text(40,line,"No of bits (16/32):");
i=lir_get_integer(60,line, 2, 16, 32);
if(kill_all_flag) return;
clear_screen();
if(i==16)
  {
  ui.input_mode=0;
  }
else
  {
  if(i==32)  
    {
    ui.input_mode=DWORD_INPUT;
    }
  else
    {  
    goto get_rxbits;
    }
  }    
// Check whether the device(s) can do whatever the user asked for.
dev=ui.ad_device_no&255;
fmt.wFormatTag=WAVE_FORMAT_PCM;
k=dev_channels[dev];
if(k>2)k=2;
fmt.nChannels=k;
fmt.nSamplesPerSec=ui.input_speed;
fmt.wBitsPerSample=i;
fmt.nBlockAlign=fmt.nChannels*fmt.wBitsPerSample/8;
fmt.nAvgBytesPerSec=fmt.nBlockAlign*fmt.nSamplesPerSec;
fmt.cbSize=0;
mmrs=waveInOpen(&hwav_rxadin1, dev, &fmt, 0, 0,
                WAVE_FORMAT_QUERY|WAVE_FORMAT_DIRECT);
if(mmrs != MMSYSERR_NOERROR)
  {
  if(mmrs == MMSYSERR_ALLOCATED)lirerr(1218);
  if(mmrs == MMSYSERR_BADDEVICEID)lirerr(1219);
  if(mmrs == MMSYSERR_NODRIVER)lirerr(1220);
  if(mmrs == MMSYSERR_NOMEM)lirerr(1221);
  if(mmrs == WAVERR_BADFORMAT)lirerr(1222);
  goto get_rxinrate;
  }
if(ui.ad_device_no>255)
  {
  dev=(ui.ad_device_no>>8)-1;
  mmrs=waveInOpen(&hwav_rxadin1, dev, &fmt, 0, 0,
                WAVE_FORMAT_QUERY|WAVE_FORMAT_DIRECT);
  if(mmrs != MMSYSERR_NOERROR)
    {
    lir_text(0,line+2,
           "ERROR  Format not supported by second device. (Press any key)");
    await_keyboard();
    if(kill_all_flag) return;
    clear_screen();
    goto get_rxinrate;
    }
  }
// If we have reached this far, the user has given correct info about
// speed and no of bits.
// Ask for the desired processing mode.
clear_screen();
sel_mode:;
lir_text(0,10,"Select radio interface:");
if(j == 2)j=3;
for(i=0; i<j; i++)
  {
  sprintf(s,"%d: %s",i+1,audiomode_text[i]);
  lir_text(5,12+i,s);
  }
line=14+i;  
chsel:;
await_processed_keyboard();
if(lir_inkey-'0'>j)goto chsel;
switch (lir_inkey)
  {
  case '1':
  ui.rx_channels=1;
  ui.ad_channels=1;
  break;

  case '2':
  ui.input_mode|=IQ_DATA;
  ui.rx_channels=1;
  ui.ad_channels=2;
  break;
    
  case '3':
  ui.input_mode|=TWO_CHANNELS;
  ui.rx_channels=2;
  ui.ad_channels=2;
  break;

  case '4':
  ui.input_mode|=TWO_CHANNELS+IQ_DATA;
  ui.rx_channels=2;
  ui.ad_channels=4;
  break;

  default:
  goto chsel;
  }
// Check that we really can do what the user wants.
// He may have selected mono for a device that can only do stereo.
// (Or at least that could be done under Linux...........)
if(ui.ad_channels == 1)
  {
// Silently ignore a second device if selected.  
  ui.ad_device_no&=255;
  if(dev_channels[dev] != 1)
    {
// Check if a stereo device can do mono.
    fmt.nChannels=1;
    mmrs=waveInOpen(&hwav_rxadin1, ui.ad_device_no, &fmt, 0, 0,
                                WAVE_FORMAT_QUERY|WAVE_FORMAT_DIRECT);
    if(mmrs != MMSYSERR_NOERROR)
      {
monerr:;
      lir_text(14,14+line,"Stereo device does not support mono");
      lir_text(14,16+line,press_any_key);
      await_keyboard();
      if(kill_all_flag) return;
      clear_screen();
      goto sel_mode;
      }
    }  
  }
if(ui.ad_channels == 2)
  {
  if( (dev_channels[dev] & 2) != 0)
    {
    ui.ad_device_no&=255;
    }
  else
    {
    if(ui.ad_device_no<256)lirerr(562921);
    dev=ui.ad_device_no&255;
    fmt.nChannels=1;
    mmrs=waveInOpen(&hwav_rxadin1, dev, &fmt, 0, 0,
                                WAVE_FORMAT_QUERY|WAVE_FORMAT_DIRECT);
    if(mmrs != MMSYSERR_NOERROR)
      {
      goto monerr;
      }
    dev=(ui.ad_device_no>>8)-1;
    mmrs=waveInOpen(&hwav_rxadin1, dev, &fmt, 0, 0,
                                WAVE_FORMAT_QUERY|WAVE_FORMAT_DIRECT);
    if(mmrs != MMSYSERR_NOERROR)
      {
      goto monerr;
      }
    }
  }   
// *************************************************************************
// Input seems OK.
// Open inputs, but do not start recording.
// This way we can (probably) avoid problems with old soundcards
// that do not have duplex capabilities.
set_sndout:;
clear_screen();
open_rx_sndin();
no_of_dadev=waveOutGetNumDevs();
if(no_of_dadev == 0)
  {
  lir_text(5,2,"No output device detected");
  if(ui.ad_device_no < SPECIFIC_DEVICE_CODES)
                                   lir_text(5,28,"while input is open.");
  lir_text(5,3,"PRESS ANY KEY");
  await_keyboard();
  return;
  }
settextcolor(14);
lir_text(10,0,"Select device for Rx output");
settextcolor(7);
line=2;
if(no_of_dadev > MAX_WAV_DEVICES)no_of_dadev=MAX_WAV_DEVICES;
for(device_id=0; device_id<no_of_dadev; device_id++)
  {
  mmrs=waveOutGetDevCaps(device_id, &pwoc, sizeof(WAVEOUTCAPS));   
  if(mmrs==MMSYSERR_NOERROR)
    {
    i=0;
    if( (pwoc.dwFormats & 
           (WAVE_FORMAT_1M16|WAVE_FORMAT_2M16|WAVE_FORMAT_4M16|
            WAVE_FORMAT_1M08|WAVE_FORMAT_2M08|WAVE_FORMAT_4M08)) != 0)i=1;
    if( (pwoc.dwFormats & 
           (WAVE_FORMAT_1S16|WAVE_FORMAT_2S16|WAVE_FORMAT_4S16|
            WAVE_FORMAT_1S08|WAVE_FORMAT_2S08|WAVE_FORMAT_4S08)) != 0)i|=2;
    dev_channels[device_id]=i;
    sprintf(s,"%d  Fmts=%4lx   %s", device_id, pwoc.dwFormats, pwoc.szPname);
    lir_text(10,line,s);
    line++;
    }
  else
    {
    dev_channels[device_id]=-1;
    }      
  }
line+=2;  
if(no_of_dadev == 1)
  {
  ui.da_device_no=0;
  j=dev_channels[0];
  if(j>2)j=2;
  }
else  
  {
  lir_text(0,line,"Select device for Rx output by line number:");
get_dadev:;
  ui.da_device_no=lir_get_integer(44,line, 2, 0, no_of_dadev-1);
  if(kill_all_flag) return;
  if(dev_channels[ui.ad_device_no]<0)goto get_dadev;
  j=dev_channels[ui.da_device_no];
  line+=2;
  }
if(j==0 || j==3)
  {
  ui.max_output_channels=2;
  ui.min_output_channels=1;
  }
else
  {
  ui.max_output_channels=j;
  ui.min_output_channels=j;
  }
mmrs=waveOutGetDevCaps(ui.da_device_no, &pwoc, sizeof(WAVEOUTCAPS));   
i=0;
if( (pwoc.dwFormats & 
           (WAVE_FORMAT_1M08|WAVE_FORMAT_2M08|WAVE_FORMAT_4M08|
            WAVE_FORMAT_1S08|WAVE_FORMAT_2S08|WAVE_FORMAT_4S08)) != 0)i=1;
if( (pwoc.dwFormats & 
           (WAVE_FORMAT_1M16|WAVE_FORMAT_2M16|WAVE_FORMAT_4M16|
            WAVE_FORMAT_1S16|WAVE_FORMAT_2S16|WAVE_FORMAT_4S16)) != 0)i|=2;
if(i==0 || i==3)
  {
  ui.max_output_bytes=2;
  ui.min_output_bytes=1;
  }
else
  {
  ui.max_output_bytes=j;
  ui.min_output_bytes=j;
  }
ui.min_output_speed=5000;
ui.max_output_speed=96000;
close_rx_sndin();
sprintf(s,"%s selected for Rx audio output.", pwoc.szPname);
lir_text(10,line,s);
sprintf(s,"Do not forget to save with 'W' in the main menu (%s)",
                                                            press_any_key);
lir_text(10,line+1,s);
await_keyboard();
}


void open_rx_sndout(void)
{
int errcod;
WAVEFORMATEX fmt;
MMRESULT mmrs;
int i, k;
if(wave_out_open_flag)return;
fmt.wFormatTag=WAVE_FORMAT_PCM;
fmt.nSamplesPerSec=genparm[DA_OUTPUT_SPEED];
fmt.wBitsPerSample=8*da_output_bytes;
fmt.cbSize=0;
fmt.nChannels=da_output_channels;
fmt.nBlockAlign=fmt.nChannels*fmt.wBitsPerSample/8;
fmt.nAvgBytesPerSec=fmt.nBlockAlign*fmt.nSamplesPerSec;
mmrs=waveOutOpen(&hwav_rxdaout, ui.da_device_no, 
                                &fmt, (DWORD)rx_daout, 0, CALLBACK_FUNCTION);
if(mmrs != MMSYSERR_NOERROR)
  {
  lirerr(1234);
  return;
  }
wave_out_open_flag=TRUE;
mmrs=waveOutPause(hwav_rxdaout);
if(mmrs != MMSYSERR_NOERROR)
  {
  lirerr(1238);
  return;
  }
errcod=1237;
waveout_buf=malloc(NO_OF_RX_WAVEOUT*daout_block+fmt.nBlockAlign);
if(waveout_buf==NULL)goto wave_close;
wave_outhdr=malloc(NO_OF_RX_WAVEOUT*sizeof(WAVEHDR));
if(wave_outhdr==NULL)goto free_waveout;
da_wrbuf=malloc(daout_block);
if(da_wrbuf==NULL)goto free_outhdr;
rxdaout_newbuf_ptr=0;
for(i=0; i<NO_OF_RX_WAVEOUT; i++)
  {
  k=(int)(waveout_buf+i*daout_block+fmt.nBlockAlign);
  k&=(-1-fmt.nBlockAlign);
  wave_outhdr[i].lpData=(char*)(k);
  wave_outhdr[i].dwBufferLength=daout_block;
  wave_outhdr[i].dwBytesRecorded=0;
  wave_outhdr[i].dwUser=1; 
  wave_outhdr[i].dwFlags=0;
  wave_outhdr[i].dwLoops=0;
  wave_outhdr[i].lpNext=NULL;
  wave_outhdr[i].reserved=0;
  mmrs=waveOutPrepareHeader(hwav_rxdaout, &wave_outhdr[i], sizeof(WAVEHDR));
  rxdaout_newbuf[i]=&wave_outhdr[i];
  if(mmrs != MMSYSERR_NOERROR)goto errfree;
  }
da_totbytes=NO_OF_RX_WAVEOUT*daout_block;
audio_out=0;
return;
errfree:;
free(da_wrbuf);
free_outhdr:;
free(wave_outhdr);
free_waveout:;
free(waveout_buf);
wave_close:;
waveOutClose(hwav_rxdaout);
wave_out_open_flag=TRUE;
lirerr(errcod);
}

